use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::ItemFn;

#[proc_macro_attribute]
pub fn testdep(_attr: TokenStream, item: TokenStream) -> TokenStream {
    let mut test: ItemFn = syn::parse(item).unwrap();

    let test_name = test.sig.ident.clone();

    let test_once = format_ident!("{}_once", test_name);
    let test_impl_name = format_ident!("{}_impl", test_name);
    let test_result = format_ident!("{}_result", test_name);
    let test_wrapper = format_ident!("{}_wrapper", test_name);

    let test_type = match test.sig.output {
        syn::ReturnType::Default => quote!(()),
        syn::ReturnType::Type(_, ref t) => quote!(#t),
    };

    let args = test.sig.inputs.iter().map(|arg| {
        let pat = match arg {
            syn::FnArg::Typed(pat) => pat,
            _ => panic!("cannot take self as an argument to a test"),
        };

        let p = &pat.pat;
        let wrapper: syn::Ident = syn::parse_str(&format!("{}_wrapper", quote!(#p))).unwrap();
        let result: syn::Ident = syn::parse_str(&format!("{}_result", quote!(#p))).unwrap();

        let access = if matches!(*pat.ty, syn::Type::Reference(_)) {
            quote!(as_ref)
        } else {
            quote!(take)
        };
        quote!({
            #wrapper();
            unsafe { #result.#access().expect("Another test has taken ownership of this item") }
        })
    });

    test.sig.ident = test_impl_name.clone();
    let test_impl = test.clone();

    let drop_on_shutdown = format_ident!("{}_drop", test_name);

    let t = quote! {
        #test_impl

        #[allow(non_upper_case_globals)]
        static #test_once: std::sync::Once = std::sync::Once::new();
        #[allow(non_upper_case_globals)]
        static mut #test_result: Option<#test_type> = None;

        #[::testdep::ctor::dtor]
        fn #drop_on_shutdown() {
            unsafe { #test_result.take(); }
        }

        fn #test_wrapper() {
            #test_once.call_once(|| {
                let result = #test_impl_name(#(#args),*);
                unsafe { #test_result = Some(result); }
            });
        }

        #[test]
        fn #test_name() {
            #test_wrapper();
        }
    };

    t.into()
}
